#ifndef INCLUDED_ATRSOFT_PATTERN_INCLUDE_LINIE18VISITOR_H
#define INCLUDED_ATRSOFT_PATTERN_INCLUDE_LINIE18VISITOR_H

/* Copyright (C) 2001 ATRSoft GmbH 
   This file is part of the ATRSoft GmbH utilitys.

   The ATRSoft GmbH utilitys is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or any later version.

   The ATRSoft GmbH utilitys is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the ATRSoft GmbH utilitys; see the file LICENSE.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

// http://www.atrsoft.de
// mailto: info@atrsoft.de

// $Id: linie18visitor.h,v 1.6 2001/10/10 08:59:53 ar Exp $

// choose one version
// single for one acceptor only
// multi for multiple acceptors

#undef ATRSOFTVISITORACCEPTORINHERITANCEMODE
#ifdef ATRSOFTMULTIVISITORACCEPTOR
#define ATRSOFTVISITORACCEPTORINHERITANCEMODE virtual
#define ATRSOFTINHERITANCESIZE 2
#else
#define ATRSOFTVISITORACCEPTORINHERITANCEMODE
#define ATRSOFTINHERITANCESIZE 1
#endif

// if your compiler is not able to instantiate the accep methods self
// you have to use the explicit specialization for accept.
// to do this, you have to compile with the following
// define. this is needed for vc 6, gnu gcc 3 needs it not ...
// #define ATRSOFTLINIE18ACCEPTOR_USE_EXPLICIT_INSTANCE


namespace ATRSoft {

namespace Pattern {

// Linie18Visitor : the class of visitor
// MyVisitorsBaseType : the type from whom all visitors will be derived.
// this is a class which is derived from Linie18Visitor itself...
// RetType : the return type of the visit and accept methods
// derive visitor base class from this class and define all visits in there
// derive visitors form visitor base class and implement visits there

template < typename MyVisitorsBaseType, typename RetType >
struct Linie18Visitor
{
    // return value of visit and accept for this visitors
    typedef RetType ReturnType;
    // type of visitor base class
    typedef MyVisitorsBaseType VisitorsBaseType;
 
    // for tests
    typedef char InheritanceMode [ATRSOFTINHERITANCESIZE];


   // We want to derive, so this should be normally needed
    // we will never throw in case of dtor calls ....
    virtual ~Linie18Visitor () throw () { }
};


// VisitorAcceptor : the proxy class for derived classes for the user.
// T1 : visited class
// T2 : Base class of hierarchy
// V : visitor hierarchy base class
// this is the acceptor class , some kind of proxy for the classes of the visited hierarchy
// derive all visited classes from this class. 
template < class T1, class T2, class V >
struct VisitorAcceptor : ATRSOFTVISITORACCEPTORINHERITANCEMODE public T2
{
    // return type
    typedef typename V::ReturnType ReturnType; 

    // the visitor base class type
    typedef typename V::VisitorsBaseType VisitorsBaseType; 

    // for tests 
    typedef char InheritanceMode [ATRSOFTINHERITANCESIZE];

    // the heart of the this stuff...
    // We are a proxy for T1. we accept the visitor (actually the type only part)
    // and call the visitor base classes appropriate visit method for us
    // the visit then casts down the this pointer to the appropriate derived class
    // whohw ..... 

#ifndef ATRSOFTLINIE18ACCEPTOR_USE_EXPLICIT_INSTANCE
    ReturnType 
    accept (V* v)
	{
	   return static_cast<VisitorsBaseType *>(v)->visit (this);
	}
#else
    ReturnType 
    accept (V* v);
#endif
};

// use this macro if your compiler cannot handle accept by itself.
// eg VC 6 needs this....
// eg gnu gcc 3 needs it not...
// set the switch in your makefile/project ...
// you need the macro normaly where you instantiate a derived class object ...
#ifdef ATRSOFTLINIE18ACCEPTOR_USE_EXPLICIT_INSTANCE
#define ATRSOFTLINIE18ACCEPTOR(__T1,__T2,__V)  \
     template <>  \
     __V::ReturnType  \
     ATRSoft::Pattern::VisitorAcceptor < __T1, __T2, __V > ::accept (__V* v)  \
     {  \
	static_cast<VisitorsBaseType *>(v)->visit (this);  \
     }
#else
#define ATRSOFTLINIE18ACCEPTOR(__T1,__T2,__V)
#endif

} // pattern

} // atrsoft

#endif

// end of file

